Package: purrrlyr
Title: Tools at the Intersection of 'purrr' and 'dplyr'
Version: 0.0.10
Authors@R: 
    c(person(given = "Lionel",
             family = "Henry",
             role = c("aut", "cre"),
             email = "lionel@posit.co"),
      person(given = "Hadley",
             family = "Wickham",
             role = "ctb",
             email = "hadley@posit.co"),
      person("Posit Software, PBC", role = c("cph", "fnd"),
             comment = c(ROR = "03wc8by49")))
Description: Some functions at the intersection of 'dplyr' and
    'purrr' that formerly lived in 'purrr'.
License: GPL-3 | file LICENSE
URL: https://github.com/hadley/purrrlyr
BugReports: https://github.com/hadley/purrrlyr/issues
Imports: dplyr (>= 0.8.0), magrittr (>= 1.5), purrr (>= 0.2.2), Rcpp
Suggests: covr, testthat (>= 3.0.0)
LinkingTo: Rcpp
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
Config/build/compilation-database: true
NeedsCompilation: yes
Packaged: 2025-09-11 15:04:28 UTC; lionel
Author: Lionel Henry [aut, cre],
  Hadley Wickham [ctb],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Lionel Henry <lionel@posit.co>
Repository: CRAN
Date/Publication: 2025-09-12 05:30:11 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 12:07:10 UTC; unix
Archs: purrrlyr.so.dSYM

# Generated by roxygen2: do not edit by hand

export("%>%")
export(by_row)
export(by_slice)
export(dmap)
export(dmap_at)
export(dmap_if)
export(invoke_rows)
export(map_rows)
export(slice_rows)
export(unslice)
importFrom(Rcpp,sourceCpp)
importFrom(dplyr,group_data)
importFrom(magrittr,"%>%")
useDynLib(purrrlyr, .registration = TRUE)

# purrrlyr

<!-- badges: start -->
[![Lifecycle: superseded](https://img.shields.io/badge/lifecycle-superseded-blue.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/purrrlyr)](https://cran.r-project.org/package=purrrlyr)
[![R-CMD-check](https://github.com/hadley/purrrlyr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/hadley/purrrlyr/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

purrrlyr contains some functions that lie at the intersection of purrr
and dplyr. They have been removed from purrr in order to make the
package lighter and because they have been replaced by other solutions
in the tidyverse.

Please see Jenny Brian's
[webinar on row-oriented workflows](https://github.com/jennybc/row-oriented-workflows#readme)
for some alternative approaches.

#' @importFrom dplyr group_data
group_labels <- function(data) {
  dplyr::select(group_data(data), -".rows")
}

group_sizes <- function(data) {
  lengths(group_data(data)$.rows)
}

subset_slices <- function(data, keep_groups = FALSE) {
  if (!dplyr::is_grouped_df(data)) {
    return(list(data))
  }

  cols <- setdiff(names(data), dplyr::group_vars(data))

  indices <- group_data(data)$.rows
  lapply(indices, function(x) data[x, cols])
}

#' Apply a function to slices of a data frame
#'
#' `by_slice()` applies `..f` on each group of a data
#' frame. Groups should be set with `slice_rows()` or
#' [dplyr::group_by()].
#'
#' `by_slice()` provides equivalent functionality to dplyr's
#' [dplyr::do()] function. In combination with
#' `map()`, `by_slice()` is equivalent to
#' [dplyr::summarise_each()] and
#' [dplyr::mutate_each()]. The distinction between
#' mutating and summarising operations is not as important as in dplyr
#' because we do not act on the columns separately. The only
#' constraint is that the mapped function must return the same number
#' of rows for each variable mapped on.
#' @param .d A sliced data frame.
#' @param ..f A function to apply to each slice. If `..f` does
#'   not return a data frame or an atomic vector, a list-column is
#'   created under the name `.out`. If it returns a data frame, it
#'   should have the same number of rows within groups and the same
#'   number of columns between groups.
#' @param ... Further arguments passed to `..f`.
#' @param .collate If "list", the results are returned as a list-
#'   column. Alternatively, if the results are data frames or atomic
#'   vectors, you can collate on "cols" or on "rows". Column collation
#'   require vector of equal length or data frames with same number of
#'   rows.
#' @param .to Name of output column.
#' @param .labels If `TRUE`, the returned data frame is prepended
#'   with the labels of the slices (the columns in `.d` used to
#'   define the slices). They are recycled to match the output size in
#'   each slice if necessary.
#' @return A data frame.
#' @seealso [by_row()], [slice_rows()],
#'   [dmap()]
#' @importFrom Rcpp sourceCpp
#' @export
#' @examples
#' # Here we fit a regression model inside each slice defined by the
#' # unique values of the column "cyl". The fitted models are returned
#' # in a list-column.
#' mtcars %>%
#'   slice_rows("cyl") %>%
#'   by_slice(purrr::partial(lm, mpg ~ disp))
#'
#' # by_slice() is especially useful in combination with map().
#'
#' # To modify the contents of a data frame, use rows collation. Note
#' # that unlike dplyr, Mutating and summarising operations can be
#' # used indistinctly.
#'
#' # Mutating operation:
#' df <- mtcars %>% slice_rows(c("cyl", "am"))
#' df %>% by_slice(dmap, ~ .x / sum(.x), .collate = "rows")
#'
#' # Summarising operation:
#' df %>% by_slice(dmap, mean, .collate = "rows")
#'
#' # Note that mapping columns within slices is best handled by dmap():
#' df %>% dmap(~ .x / sum(.x))
#' df %>% dmap(mean)
#'
#' # If you don't need the slicing variables as identifiers, switch
#' # .labels to FALSE:
#' mtcars %>%
#'   slice_rows("cyl") %>%
#'   by_slice(purrr::partial(lm, mpg ~ disp), .labels = FALSE) %>%
#'   purrr::flatten() %>%
#'   purrr::map(coef)
by_slice <- function(.d, ..f, ..., .collate = c("list", "rows", "cols"),
                     .to = ".out", .labels = TRUE) {
  deprecate("by_slice() is deprecated. Please use the new colwise family in dplyr.\n",
    "E.g., summarise_all(), mutate_all(), etc.")
  ..f <- as_rows_function(..f)
  if (!dplyr::is.grouped_df(.d)) {
    stop(".d must be a sliced data frame", call. = FALSE)
  }
  if (length(.d) <= length(group_labels(.d))) {
    stop("Mappable part of data frame is empty", call. = FALSE)
  }
  .collate <- match.arg(.collate)

  set_sliced_env(.d, .labels, .collate, .to, environment(), ".d")
  env <- environment()
  env$.d <- subset_slices(.d)
  .Call(by_slice_impl, env, ".d", "..f")
}

# Prevents as_function() from transforming to a plucking function
as_rows_function <- function(f, f_name = ".f") {
  if (inherits(f, "formula")) {
    as_function(f)
  } else if (!is.function(f)) {
    stop(f_name, " should be a function or a formula", call. = FALSE)
  } else {
    f
  }
}

set_sliced_env <- function(df, labels, collate, to, env, x_name) {
  env$.unique_labels <- TRUE
  env$.labels <- labels;
  env$.collate <- collate
  env$.to <- to
  env$.labels_cols <- group_labels(df)
  env$.slicing_cols <- df[names(env$.labels_cols) %||% character(0)]
}


#' Apply a function to each row of a data frame
#'
#' `by_row()` and `invoke_rows()` apply `..f` to each row
#' of `.d`. If `..f`'s output is not a data frame nor an
#' atomic vector, a list-column is created. In all cases,
#' `by_row()` and `invoke_rows()` create a data frame in tidy
#' format.
#'
#' By default, the whole row is appended to the result to serve as
#' identifier (set `.labels` to `FALSE` to prevent this). In
#' addition, if `..f` returns a multi-rows data frame or a
#' non-scalar atomic vector, a `.row` column is appended to
#' identify the row number in the original data frame.
#'
#' `invoke_rows()` is intended to provide a version of
#' `pmap()` for data frames. Its default collation method is
#' `"cols"`, which makes it equivalent to
#' `mdply()` from the plyr package. Note that
#' `invoke_rows()` follows the signature pattern of the
#' `invoke` family of functions and takes `.f` as its first
#' argument.
#'
#' The distinction between `by_row()` and `invoke_rows()` is
#' that the former passes a data frame to `..f` while the latter
#' maps the columns to its function call. This is essentially like
#' using `invoke()` with each row. Another way to view
#' this is that `invoke_rows()` is equivalent to using
#' `by_row()` with a function lifted to accept dots (see
#' `lift()`).
#'
#' @param .d A data frame.
#' @param .f,..f A function to apply to each row. If `..f` does
#'   not return a data frame or an atomic vector, a list-column is
#'   created under the name `.out`. If it returns a data frame, it
#'   should have the same number of rows within groups and the same
#'   number of columns between groups.
#' @param ... Further arguments passed to `..f`.
#' @inheritParams by_slice
#' @return A data frame.
#' @seealso [by_slice()]
#' @export
#' @examples
#' # ..f should be able to work with a list or a data frame. As it
#' # happens, sum() handles data frame so the following works:
#' mtcars %>% by_row(sum)
#'
#' # Other functions such as mean() may need to be adjusted with one
#' # of the lift_xy() helpers:
#' mtcars %>% by_row(purrr::lift_vl(mean))
#'
#' # To run a function with invoke_rows(), make sure it is variadic (that
#' # it accepts dots) or that .f's signature is compatible with the
#' # column names
#' mtcars %>% invoke_rows(.f = sum)
#' mtcars %>% invoke_rows(.f = purrr::lift_vd(mean))
#'
#' # invoke_rows() with cols collation is equivalent to plyr::mdply()
#' p <- expand.grid(mean = 1:5, sd = seq(0, 1, length = 10))
#' p %>% invoke_rows(.f = rnorm, n = 5, .collate = "cols")
#' \dontrun{
#' p %>% plyr::mdply(rnorm, n = 5) %>% dplyr::tbl_df()
#' }
#'
#' # To integrate the result as part of the data frame, use rows or
#' # cols collation:
#' mtcars[1:2] %>% by_row(function(x) 1:5)
#' mtcars[1:2] %>% by_row(function(x) 1:5, .collate = "rows")
#' mtcars[1:2] %>% by_row(function(x) 1:5, .collate = "cols")
by_row <- function(.d, ..f, ..., .collate = c("list", "rows", "cols"),
                   .to = ".out", .labels = TRUE) {
  deprecate("`by_row()` is deprecated; please use a combination of:\n",
    "tidyr::nest(); dplyr::mutate(); purrr::map()")
  check_df_consistency(.d)
  if (nrow(.d) < 1) {
    return(.d)
  }

  ..f <- as_rows_function(..f)
  .collate <- match.arg(.collate)

  .unique_labels <- 0
  .labels_cols <- .d
  .slicing_cols <- .d

  .d <- lapply(seq_len(nrow(.d)), function(i) .d[i, , drop = FALSE])

  .Call(by_slice_impl, environment(), ".d", "..f")
}

check_df_consistency <- function(.d) {
  if (!is.data.frame(.d)) {
    stop(".d must be a data frame", call. = FALSE)
  }
  if (length(.d) == 0) {
    stop("Data frame is empty", call. = FALSE)
  }
}

#' @rdname by_row
#' @export
invoke_rows <- function(.f, .d, ..., .collate = c("list", "rows", "cols"),
                        .to = ".out", .labels = TRUE) {
  deprecate("`invoke_rows()` is deprecated; please use `pmap()` instead.")
  check_df_consistency(.d)
  .collate <- match.arg(.collate)

  .unique_labels <- 0
  .labels_cols <- .d
  .slicing_cols <- .d

  .Call(invoke_rows_impl, environment(), ".d", ".f")
}

#' @export
#' @usage NULL
#' @rdname by_row
map_rows <- function(.d, .f, ..., .labels = TRUE) {
  deprecate("`map_rows()` is deprecated; please use `pmap()` instead.")
  invoke_rows(.f, .d, ..., .labels = .labels)
}


#' Slice a data frame into groups of rows
#'
#' `slice_rows()` is equivalent to dplyr's
#' [dplyr::group_by()] command but it takes a vector of
#' column names or positions instead of capturing column names with
#' special evaluation. `unslice()` removes the slicing
#' attributes.
#' @param .d A data frame to slice or unslice.
#' @param .cols A character vector of column names or a numeric vector
#'   of column positions. If `NULL`, the slicing attributes are
#'   removed.
#' @return A sliced or unsliced data frame.
#' @seealso [by_slice()] and [dplyr::group_by()]
#' @export
slice_rows <- function(.d, .cols = NULL) {
  deprecate("`slice_rows()` is deprecated; please use `dplyr::group_by()` instead.")
  stopifnot(is.data.frame(.d))
  if (is.null(.cols)) {
    return(unslice(.d))
  }
  if (is.numeric(.cols)) {
    .cols <- names(.d)[.cols]
  }
  stopifnot(is.character(.cols))

  dplyr::group_by_at(.d, .cols)
}

#' @rdname slice_rows
#' @export
unslice <- function(.d) {
  deprecate("`unslice()` is deprecated; please use `dplyr::ungroup()` instead.")
  dplyr::ungroup(.d)
}

test_that("output column is named according to .to", {
  output1 <- mtcars %>% slice_rows("cyl") %>% by_slice(~ list(NULL), .to = "my_col", .labels = FALSE)
  output2 <- mtcars %>% by_row(~ list(NULL), .to = "my_col", .labels = FALSE)
  output3 <- mtcars %>% invoke_rows(.f = function(...) list(NULL), .collate = "list", .to = "my_col", .labels = FALSE)

  expect_equal(names(output1), "my_col")
  expect_equal(names(output2), "my_col")
  expect_equal(names(output3), "my_col")
})

test_that("empty", {
  rows_collation <- invoke_rows(empty, mtcars[1:2], .collate = "rows")
  cols_collation <- invoke_rows(empty, mtcars[1:2], .collate = "cols")
  list_collation <- invoke_rows(empty, mtcars[1:2], .collate = "list")

  expect_equal(rows_collation$.out, numeric(0))
  expect_equal(cols_collation$.out, numeric(0))
  expect_equal(list_collation$.out, purrr::rerun(32, numeric(0)))

  expect_equal(dim(rows_collation), c(0, 3))
  expect_equal(dim(cols_collation), c(0, 3))
  expect_equal(dim(list_collation), c(32, 3))
})

test_that("all nulls fail, except with list-collation", {
  expect_error(invoke_rows(all_nulls, mtcars[1:2], .collate = "rows"))
  expect_error(invoke_rows(all_nulls, mtcars[1:2], .collate = "cols"))

  list_collation <- invoke_rows(all_nulls, mtcars[1:2], .collate = "list")
  expect_equal(list_collation$.out, vector("list", 32))
  expect_equal(dim(list_collation), c(32, 3))
})

test_that("scalars", {
  rows_collation <- invoke_rows(scalars, mtcars[1:2], .collate = "rows")
  cols_collation <- invoke_rows(scalars, mtcars[1:2], .collate = "cols")
  list_collation <- invoke_rows(scalars, mtcars[1:2], .collate = "list")

  out <- paste("a", mtcars$mpg)

  expect_equal(rows_collation$.out, out)
  expect_equal(cols_collation$.out, out)
  expect_equal(list_collation$.out, as.list(out))

  expect_equal(dim(rows_collation), c(32, 3))
  expect_equal(dim(cols_collation), c(32, 3))
  expect_equal(dim(list_collation), c(32, 3))
})

test_that("scalars with some nulls", {
  rows_collation <- invoke_rows(scalar_nulls, mtcars[1:2], .collate = "rows")
  cols_collation <- invoke_rows(scalar_nulls, mtcars[1:2], .collate = "cols")
  list_collation <- invoke_rows(scalar_nulls, mtcars[1:2], .collate = "list")

  expect_equal(rows_collation$.out, rep(1, 16))
  expect_equal(cols_collation$.out, rep(1, 16))
  expect_equal(list_collation$.out, rep(list(1L, NULL), 16))

  expect_equal(dim(rows_collation), c(16, 3))
  expect_equal(dim(cols_collation), c(16, 3))
  expect_equal(dim(list_collation), c(32, 3))

  # Make sure properties are well inferred when first result is NULL
  rows_collation <- invoke_rows(scalar_first_nulls, mtcars[1:2], .collate = "rows")
  expect_equal(rows_collation$.out, rep(1, 16))
})

test_that("labels are correctly subsetted", {
  rows_collation <- invoke_rows(scalar_first_nulls, mtcars[1:2], .collate = "rows")
  expect_equal(rows_collation[1:2], dplyr::as_tibble(mtcars[seq(2, 32, 2), 1:2]))
})

test_that("vectors", {
  rows_collation <- invoke_rows(vectors, mtcars[1:2], .collate = "rows")
  cols_collation <- invoke_rows(vectors, mtcars[1:2], .collate = "cols")
  list_collation <- invoke_rows(vectors, mtcars[1:2], .collate = "list")

  data <- dplyr::rowwise(mtcars[1:2])
  out <- dplyr::do(data, .out = paste(c("a", "b"), c(.$mpg, .$cyl)))[[1]]

  expect_equal(rows_collation$.row, rep(1:32, each = 2))
  expect_equal(rows_collation$.out, unlist(out))

  expect_equal(cols_collation$.out1, paste("a", mtcars$mpg))
  expect_equal(cols_collation$.out2, paste("b", mtcars$cyl))
  expect_equal(list_collation$.out, out)

  expect_equal(dim(rows_collation), c(64, 4))
  expect_equal(dim(cols_collation), c(32, 4))
  expect_equal(dim(list_collation), c(32, 3))
})

test_that("data frames", {
  rows_collation <- invoke_rows(dataframes, mtcars[1:2], .collate = "rows")
  cols_collation <- invoke_rows(dataframes, mtcars[1:2], .collate = "cols")
  list_collation <- invoke_rows(dataframes, mtcars[1:2], .collate = "list")

  expect_equal(rows_collation$.row, rep(1:32, each = 3))
  expect_equal(rows_collation[4:5], dplyr::as_tibble(dplyr::bind_rows(purrr::rerun(32, df))))
  expect_equal(cols_collation[[3]], rep(df[[1]][1], 32))
  expect_equal(cols_collation[[8]], rep(df[[2]][3], 32))
  expect_equal(list_collation$.out, purrr::rerun(32, df))

  expect_equal(dim(rows_collation), c(96, 5))
  expect_equal(dim(cols_collation), c(32, 8))
  expect_equal(dim(list_collation), c(32, 3))
})

test_that("data frames with some nulls/empty", {
  rows_collation <- invoke_rows(dataframes_nulls, mtcars[1:2], .collate = "rows")
  cols_collation <- invoke_rows(dataframes_nulls, mtcars[1:2], .collate = "cols")
  list_collation <- invoke_rows(dataframes_nulls, mtcars[1:2], .collate = "list")

  expect_equal(rows_collation[4:5], dplyr::as_tibble(dplyr::bind_rows(purrr::rerun(16, df))))
  expect_equal(list_collation$.out, rep(list(df, NULL), 16))

  expect_equal(dim(rows_collation), c(48, 5))
  expect_equal(dim(cols_collation), c(16, 8))
  expect_equal(dim(list_collation), c(32, 3))
})

test_that("empty data frames", {
  rows_collation_by_row <- invoke_rows(empty_dataframes, mtcars[1:2], .collate = "rows")
  rows_collation_by_slice <- by_slice(grouped, empty_dataframes, .collate = "rows")

  expect_equal(rows_collation_by_row[4:5], dplyr::as_tibble(df[0, ]))
  expect_equal(rows_collation_by_slice[2:3], dplyr::as_tibble(df[0, ]))

  expect_equal(dim(rows_collation_by_row), c(0, 5))
  expect_equal(dim(rows_collation_by_slice), c(0, 3))
})

test_that("some empty data frames", {
  rows_collation_by_row <- invoke_rows(some_empty_dataframes, mtcars[1:2], .collate = "rows")
  rows_collation_by_slice <- by_slice(grouped, some_empty_dataframes, .collate = "rows")

  expect_equal(rows_collation_by_row[4:5], dplyr::as_tibble(dplyr::bind_rows(purrr::rerun(16, df))))
  expect_equal(rows_collation_by_slice[2:3], dplyr::as_tibble(dplyr::bind_rows(purrr::rerun(2, df))))

  expect_equal(dim(rows_collation_by_row), c(48, 5))
  expect_equal(dim(rows_collation_by_slice), c(6, 3))
})

test_that("unconsistent data frames fail", {
  expect_error(invoke_rows(unconsistent_names, mtcars[1:2], .collate = "rows"), "consistent names")
  expect_error(invoke_rows(unconsistent_types, mtcars[1:2], .collate = "rows"), "must return either data frames or vectors")
})

test_that("objects", {
  list_collation <- invoke_rows(test_objects, mtcars[1:2], .collate = "list")

  expect_equal(
    list_collation$.out,
    rep(list(function() {}), 32),
    ignore_function_env = TRUE
  )

  expect_equal(dim(list_collation), c(32, 3))

  expect_error(invoke_rows(test_objects, mtcars[1:2], .collate = "rows"))
  expect_error(invoke_rows(test_objects, mtcars[1:2], .collate = "cols"))
})

test_that("collation of ragged objects on cols fails", {
  expect_error(invoke_rows(ragged_dataframes, mtcars[1:2], .collate = "cols"))
  expect_error(invoke_rows(ragged_vectors, mtcars[1:2], .collate = "cols"))
})

test_that("by_slice() works with slicers of different types", {
  df1 <- slice_rows(mtcars, "cyl")
  df2 <- dmap_at(mtcars, "cyl", as.character) %>% slice_rows("cyl")
  out1 <- by_slice(df1, purrr::map, mean)
  out2 <- by_slice(df2, purrr::map, mean)
  expect_identical(out1[-1], out2[-1])
  expect_equal(typeof(out1$cyl), "double")
  expect_equal(typeof(out2$cyl), "character")
})

test_that("by_slice() does not create .row column", {
  data <- slice_rows(mtcars[1:2], "cyl")

  rows_vectors <- by_slice(data, function(x) 1:3, .collate = "rows")
  expect_equal(dim(rows_vectors), c(9, 2))
  expect_equal(names(rows_vectors), c("cyl", ".out"))

  rows_dfs <- by_slice(data, function(x) df, .collate = "rows")
  expect_equal(dim(rows_dfs), c(9, 3))
  expect_equal(names(rows_dfs), c("cyl", "wt", "qsec"))
})

test_that("by_slice() fails with ungrouped data frames", {
  expect_error(by_slice(mtcars, list))
})

test_that("by_row() creates indices with c++ style indexing", {
  out <- mtcars[1:5, 1:2] %>% by_row(~ .$cyl[1])
  expect_equal(out$.out[[5]], 8)
})

test_that("error is thrown when no columns to map", {
  expect_error(mtcars["cyl"] %>% slice_rows("cyl") %>% by_slice(list), "empty")
  expect_error(dplyr::tibble() %>% invoke_rows(.f = c), "empty")
  expect_error(dplyr::tibble() %>% by_row(c), "empty")
})

test_that("grouping list-columns are copied (#9)", {
  df <- dplyr::tibble(x = as.list(1:2))
  exp <- dplyr::tibble(x = list(1L, 2L), .out = list(NA, NA))
  expect_identical(by_row(df, ~NA), exp)
})

#' Pipe operator
#'
#' @name %>%
#' @rdname pipe
#' @keywords internal
#' @export
#' @importFrom magrittr %>%
#' @usage lhs \%>\% rhs
NULL

names2 <- function(x) {
  names(x) %||% rep("", length(x))
}

`%||%` <- function(x, y) {
  if (is.null(x)) {
    y
  } else {
    x
  }
}

isFALSE <- function(x) identical(x, FALSE)

ndots <- function(...) nargs()

inv_which <- function(x, sel) {
  if (is.character(sel)) {
    names <- names(x)
    if (is.null(names)) {
      stop("character indexing requires a named object", call. = FALSE)
    }
    names %in% sel
  } else if (is.numeric(sel)) {
    seq_along(x) %in% sel
  } else {
    stop("unrecognised index type", call. = FALSE)
  }
}

deprecate <- function(...) {
  # No message for now
}

as_function <- function(...) {
  purrr::as_mapper(...)
}

df <- mtcars[1:3, c("wt", "qsec")]
df[[2]] <- as.character(df[[2]])
suppressWarnings(grouped <- slice_rows(mtcars[1:2], "cyl"))

gen_alternatives <- function(first, alt) {
  prev_alt <- TRUE
  function(...) {
    if (prev_alt) {
      out <- first
    } else {
      out <- alt
    }
    prev_alt <<- !prev_alt
    out
  }
}

all_nulls <- function(...) NULL
scalars <- function(...) paste("a", ..1)
empty <- function(...) numeric(0)
vectors <- function(...) paste(letters[1:2], c(...))
dataframes <- function(...) df
empty_dataframes <- function(...) df[0, ]
test_objects <- function(...) function() {}

scalar_nulls <- gen_alternatives(1L, NULL)
scalar_first_nulls <- gen_alternatives(NULL, 1L)
scalar_first_nulls <- gen_alternatives(NULL, 1L)
dataframes_nulls <- gen_alternatives(df, NULL)
some_empty_dataframes <- gen_alternatives(df, df[0, ])
unconsistent_names <- gen_alternatives(df, purrr::set_names(df, 1:2))
unconsistent_types <- gen_alternatives(df, purrr::map(df, as.character))
ragged_dataframes <- gen_alternatives(df, rbind(df, df))
ragged_vectors <- gen_alternatives(letters[1:2], rep(letters[1:2], 2))

#' Map over the columns of a data frame
#'
#' `dmap()` is just like [purrr::map()] but always returns a
#' data frame. In addition, it handles grouped or sliced data frames.
#'
#' `dmap_at()` and `dmap_if()` recycle length 1 vectors to
#' the group sizes.
#' @inheritParams purrr::map
#' @inheritParams rlang::as_function
#' @inheritParams purrr::map_if
#' @param .d A data frame.
#' @export
#' @examples
#' # dmap() always returns a data frame:
#' dmap(mtcars, summary)
#'
#' # dmap() also supports sliced data frames:
#' sliced_df <- mtcars[1:5] %>% slice_rows("cyl")
#' sliced_df %>% dmap(mean)
#' sliced_df %>% dmap(~ .x / max(.x))
#'
#' # This is equivalent to the combination of by_slice() and dmap()
#' # with 'rows' collation of results:
#' sliced_df %>% by_slice(dmap, mean, .collate = "rows")
dmap <- function(.d, .f, ...) {
  deprecate("dmap() is deprecated. Please use the new colwise family in dplyr.\n",
    "E.g., summarise_all(), mutate_all(), etc.")
  .f <- as_function(.f, ...)
  if (dplyr::is.grouped_df(.d)) {
    sliced_dmap(.d, .f, ...)
  } else {
    res <- .Call(map_impl, environment(), ".d", ".f", "list")
    dplyr::as_tibble(res)
  }
}

sliced_dmap <- function(.d, .f, ...) {
  if (length(.d) <= length(group_labels(.d))) {
    .d
  } else {
    set_sliced_env(.d, TRUE, "rows", "", environment(), ".d")
    slices <- subset_slices(.d)
    .Call(map_by_slice_impl, environment(), ".d", ".f", slices)
  }
}


#' @rdname dmap
#' @export
dmap_at <- function(.d, .at, .f, ...) {
  deprecate("dmap_at() is deprecated. Please use the new colwise family in dplyr.\n",
    "E.g., summarise_at(), mutate_at(), etc.")
  sel <- inv_which(.d, .at)
  partial_dmap(.d, sel, .f, ...)
}

#' @rdname dmap
#' @export
dmap_if <- function(.d, .p, .f, ...) {
  deprecate("dmap_if() is deprecated. Please use the new colwise family in dplyr.\n",
    "E.g., summarise_if(), mutate_if(), etc.")
  sel <- purrr::map_lgl(.d, .p)
  partial_dmap(.d, sel, .f, ...)
}

partial_dmap <- function(.d, .sel, .f, ...) {
  .f <- as_function(.f)
  subset <- dplyr::select(.d, !!dplyr::group_vars(.d), !!names(.d)[.sel])

  set_sliced_env(.d, FALSE, "rows", "", environment(), "slices")
  slices <- subset_slices(subset)
  res <- .Call(map_by_slice_impl, environment(), "slices", ".f", slices)

  res <- dmap_recycle(res, .d)
  .d[.sel] <- res

  .d
}

dmap_recycle <- function(res, d) {
  if (dplyr::is.grouped_df(d)) {
    return(dmap_recycle_sliced(res, d))
  }

  if (!nrow(res) %in% c(0, 1, nrow(d))) {
    stop("dmap() only recycles vectors of length 1", call. = TRUE)
  }

  res
}

dmap_recycle_sliced <- function(res, d) {
  if (nrow(res) == nrow(d)) {
    return(res)
  }

  if (nrow(group_labels(d)) == nrow(res)) {
    sizes <- group_sizes(d)
    indices <- purrr::map2(seq_len(nrow(res)), sizes, ~rep(.x, each = .y))
    res <- res[purrr::flatten_int(indices), ]
    return(res)
  }

  stop("dmap() only recycles vectors of length 1")
}

test_that("dmap() returns a data frame", {
  expect_s3_class(dmap(mtcars, mean), "data.frame")
})

test_that("dmap() works with sliced data frames", {
  df <- slice_rows(mtcars, "cyl")
  actual <- dmap(df, mean)
  expected <- by_slice(df, dmap, mean, .collate = "rows")
  expect_equal(actual, expected)
})

test_that("dmap() works with no columns to map", {
  res <- mtcars["cyl"] %>% slice_rows("cyl") %>% dmap(mean)
  expect_equal(res, dplyr::group_by(mtcars["cyl"], cyl))
})

test_that("dmap() recycles only vectors of length 1", {
  expect_equal(dmap_at(mtcars, "cyl", mean)$cyl, rep(mean(mtcars$cyl), nrow(mtcars)))
  expect_error(dmap_at(mtcars, c("cyl", "am"), ~1:2), "only recycles")
})

test_that("conditional sliced mapping recycles within groups", {
  df <- mtcars %>% slice_rows(c("vs", "am"))
  expected_df <- by_slice(df, dmap, mean, .collate = "rows")

  res_at <- dmap_at(df, c("disp", "drat"), mean)
  res_if <- dmap_if(df, ~ .x[[1]] == 160, mean)

  expected <- purrr::map2(expected_df$disp, group_sizes(df), rep) %>% purrr::flatten_dbl()
  expect_equal(res_at$disp, expected)
  expect_equal(res_if$disp, expected)
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rows.R
\name{by_slice}
\alias{by_slice}
\title{Apply a function to slices of a data frame}
\usage{
by_slice(.d, ..f, ..., .collate = c("list", "rows", "cols"),
  .to = ".out", .labels = TRUE)
}
\arguments{
\item{.d}{A sliced data frame.}

\item{..f}{A function to apply to each slice. If \code{..f} does
not return a data frame or an atomic vector, a list-column is
created under the name \code{.out}. If it returns a data frame, it
should have the same number of rows within groups and the same
number of columns between groups.}

\item{...}{Further arguments passed to \code{..f}.}

\item{.collate}{If "list", the results are returned as a list-
column. Alternatively, if the results are data frames or atomic
vectors, you can collate on "cols" or on "rows". Column collation
require vector of equal length or data frames with same number of
rows.}

\item{.to}{Name of output column.}

\item{.labels}{If \code{TRUE}, the returned data frame is prepended
with the labels of the slices (the columns in \code{.d} used to
define the slices). They are recycled to match the output size in
each slice if necessary.}
}
\value{
A data frame.
}
\description{
\code{by_slice()} applies \code{..f} on each group of a data
frame. Groups should be set with \code{slice_rows()} or
\code{\link[dplyr:group_by]{dplyr::group_by()}}.
}
\details{
\code{by_slice()} provides equivalent functionality to dplyr's
\code{\link[dplyr:do]{dplyr::do()}} function. In combination with
\code{map()}, \code{by_slice()} is equivalent to
\code{\link[dplyr:summarise_each]{dplyr::summarise_each()}} and
\code{\link[dplyr:mutate_each]{dplyr::mutate_each()}}. The distinction between
mutating and summarising operations is not as important as in dplyr
because we do not act on the columns separately. The only
constraint is that the mapped function must return the same number
of rows for each variable mapped on.
}
\examples{
# Here we fit a regression model inside each slice defined by the
# unique values of the column "cyl". The fitted models are returned
# in a list-column.
mtcars \%>\%
  slice_rows("cyl") \%>\%
  by_slice(purrr::partial(lm, mpg ~ disp))

# by_slice() is especially useful in combination with map().

# To modify the contents of a data frame, use rows collation. Note
# that unlike dplyr, Mutating and summarising operations can be
# used indistinctly.

# Mutating operation:
df <- mtcars \%>\% slice_rows(c("cyl", "am"))
df \%>\% by_slice(dmap, ~ .x / sum(.x), .collate = "rows")

# Summarising operation:
df \%>\% by_slice(dmap, mean, .collate = "rows")

# Note that mapping columns within slices is best handled by dmap():
df \%>\% dmap(~ .x / sum(.x))
df \%>\% dmap(mean)

# If you don't need the slicing variables as identifiers, switch
# .labels to FALSE:
mtcars \%>\%
  slice_rows("cyl") \%>\%
  by_slice(purrr::partial(lm, mpg ~ disp), .labels = FALSE) \%>\%
  purrr::flatten() \%>\%
  purrr::map(coef)
}
\seealso{
\code{\link[=by_row]{by_row()}}, \code{\link[=slice_rows]{slice_rows()}},
\code{\link[=dmap]{dmap()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rows.R
\name{by_row}
\alias{by_row}
\alias{invoke_rows}
\alias{map_rows}
\title{Apply a function to each row of a data frame}
\usage{
by_row(.d, ..f, ..., .collate = c("list", "rows", "cols"),
  .to = ".out", .labels = TRUE)

invoke_rows(.f, .d, ..., .collate = c("list", "rows", "cols"),
  .to = ".out", .labels = TRUE)
}
\arguments{
\item{.d}{A data frame.}

\item{...}{Further arguments passed to \code{..f}.}

\item{.collate}{If "list", the results are returned as a list-
column. Alternatively, if the results are data frames or atomic
vectors, you can collate on "cols" or on "rows". Column collation
require vector of equal length or data frames with same number of
rows.}

\item{.to}{Name of output column.}

\item{.labels}{If \code{TRUE}, the returned data frame is prepended
with the labels of the slices (the columns in \code{.d} used to
define the slices). They are recycled to match the output size in
each slice if necessary.}

\item{.f, ..f}{A function to apply to each row. If \code{..f} does
not return a data frame or an atomic vector, a list-column is
created under the name \code{.out}. If it returns a data frame, it
should have the same number of rows within groups and the same
number of columns between groups.}
}
\value{
A data frame.
}
\description{
\code{by_row()} and \code{invoke_rows()} apply \code{..f} to each row
of \code{.d}. If \code{..f}'s output is not a data frame nor an
atomic vector, a list-column is created. In all cases,
\code{by_row()} and \code{invoke_rows()} create a data frame in tidy
format.
}
\details{
By default, the whole row is appended to the result to serve as
identifier (set \code{.labels} to \code{FALSE} to prevent this). In
addition, if \code{..f} returns a multi-rows data frame or a
non-scalar atomic vector, a \code{.row} column is appended to
identify the row number in the original data frame.

\code{invoke_rows()} is intended to provide a version of
\code{pmap()} for data frames. Its default collation method is
\code{"cols"}, which makes it equivalent to
\code{mdply()} from the plyr package. Note that
\code{invoke_rows()} follows the signature pattern of the
\code{invoke} family of functions and takes \code{.f} as its first
argument.

The distinction between \code{by_row()} and \code{invoke_rows()} is
that the former passes a data frame to \code{..f} while the latter
maps the columns to its function call. This is essentially like
using \code{\link[=invoke]{invoke()}} with each row. Another way to view
this is that \code{invoke_rows()} is equivalent to using
\code{by_row()} with a function lifted to accept dots (see
\code{\link[=lift]{lift()}}).
}
\examples{
# ..f should be able to work with a list or a data frame. As it
# happens, sum() handles data frame so the following works:
mtcars \%>\% by_row(sum)

# Other functions such as mean() may need to be adjusted with one
# of the lift_xy() helpers:
mtcars \%>\% by_row(purrr::lift_vl(mean))

# To run a function with invoke_rows(), make sure it is variadic (that
# it accepts dots) or that .f's signature is compatible with the
# column names
mtcars \%>\% invoke_rows(.f = sum)
mtcars \%>\% invoke_rows(.f = purrr::lift_vd(mean))

# invoke_rows() with cols collation is equivalent to plyr::mdply()
p <- expand.grid(mean = 1:5, sd = seq(0, 1, length = 10))
p \%>\% invoke_rows(.f = rnorm, n = 5, .collate = "cols")
\dontrun{
p \%>\% plyr::mdply(rnorm, n = 5) \%>\% dplyr::tbl_df()
}

# To integrate the result as part of the data frame, use rows or
# cols collation:
mtcars[1:2] \%>\% by_row(function(x) 1:5)
mtcars[1:2] \%>\% by_row(function(x) 1:5, .collate = "rows")
mtcars[1:2] \%>\% by_row(function(x) 1:5, .collate = "cols")
}
\seealso{
\code{\link[=by_slice]{by_slice()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rows.R
\name{slice_rows}
\alias{slice_rows}
\alias{unslice}
\title{Slice a data frame into groups of rows}
\usage{
slice_rows(.d, .cols = NULL)

unslice(.d)
}
\arguments{
\item{.d}{A data frame to slice or unslice.}

\item{.cols}{A character vector of column names or a numeric vector
of column positions. If \code{NULL}, the slicing attributes are
removed.}
}
\value{
A sliced or unsliced data frame.
}
\description{
\code{slice_rows()} is equivalent to dplyr's
\code{\link[dplyr:group_by]{dplyr::group_by()}} command but it takes a vector of
column names or positions instead of capturing column names with
special evaluation. \code{unslice()} removes the slicing
attributes.
}
\seealso{
\code{\link[=by_slice]{by_slice()}} and \code{\link[dplyr:group_by]{dplyr::group_by()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dmap.R
\name{dmap}
\alias{dmap}
\alias{dmap_at}
\alias{dmap_if}
\title{Map over the columns of a data frame}
\usage{
dmap(.d, .f, ...)

dmap_at(.d, .at, .f, ...)

dmap_if(.d, .p, .f, ...)
}
\arguments{
\item{.d}{A data frame.}

\item{.f}{A function, specified in one of the following ways:
\itemize{
\item A named function, e.g. \code{mean}.
\item An anonymous function, e.g. \verb{\\(x) x + 1} or \code{function(x) x + 1}.
\item A formula, e.g. \code{~ .x + 1}. You must use \code{.x} to refer to the first
argument. No longer recommended.
\item A string, integer, or list, e.g. \code{"idx"}, \code{1}, or \code{list("idx", 1)} which
are shorthand for \verb{\\(x) pluck(x, "idx")}, \verb{\\(x) pluck(x, 1)}, and
\verb{\\(x) pluck(x, "idx", 1)} respectively. Optionally supply \code{.default} to
set a default value if the indexed element is \code{NULL} or does not exist.
}

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Wrap a function with \code{\link[purrr:in_parallel]{in_parallel()}} to declare that it should be performed
in parallel. See \code{\link[purrr:in_parallel]{in_parallel()}} for more details.
Use of \code{...} is not permitted in this context.}

\item{...}{Additional arguments passed on to the mapped function.

We now generally recommend against using \code{...} to pass additional
(constant) arguments to \code{.f}. Instead use a shorthand anonymous function:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{# Instead of
x |> map(f, 1, 2, collapse = ",")
# do:
x |> map(\\(x) f(x, 1, 2, collapse = ","))
}\if{html}{\out{</div>}}

This makes it easier to understand which arguments belong to which
function and will tend to yield better error messages.}

\item{.at}{A logical, integer, or character vector giving the elements
to select. Alternatively, a function that takes a vector of names,
and returns a logical, integer, or character vector of elements to select.

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}: if the tidyselect package is
installed, you can use \code{vars()} and tidyselect helpers to select
elements.}

\item{.p}{A single predicate function, a formula describing such a
predicate function, or a logical vector of the same length as \code{.x}.
Alternatively, if the elements of \code{.x} are themselves lists of
objects, a string indicating the name of a logical element in the
inner lists. Only those elements where \code{.p} evaluates to
\code{TRUE} will be modified.}
}
\description{
\code{dmap()} is just like \code{\link[purrr:map]{purrr::map()}} but always returns a
data frame. In addition, it handles grouped or sliced data frames.
}
\details{
\code{dmap_at()} and \code{dmap_if()} recycle length 1 vectors to
the group sizes.
}
\examples{
# dmap() always returns a data frame:
dmap(mtcars, summary)

# dmap() also supports sliced data frames:
sliced_df <- mtcars[1:5] \%>\% slice_rows("cyl")
sliced_df \%>\% dmap(mean)
sliced_df \%>\% dmap(~ .x / max(.x))

# This is equivalent to the combination of by_slice() and dmap()
# with 'rows' collation of results:
sliced_df \%>\% by_slice(dmap, mean, .collate = "rows")
}
